""" jmeter load generator support """
import datetime
import logging
import os
import signal
import subprocess
import time

from pkg_resources import resource_string
from ...common.util import splitstring
from ...common.interfaces import AbstractPlugin, AggregateResultListener, AbstractInfoWidget, GeneratorPlugin

from .reader import JMeterReader
from ..Aggregator import Plugin as AggregatorPlugin
from ..Console import Plugin as ConsolePlugin
from ..Console import screen as ConsoleScreen

logger = logging.getLogger(__name__)


class Plugin(AbstractPlugin, GeneratorPlugin):
    """ JMeter tank plugin """
    SECTION = 'jmeter'

    def __init__(self, core):
        AbstractPlugin.__init__(self, core)
        self.jmeter_process = None
        self.args = None
        self.original_jmx = None
        self.jtl_file = None
        self.ext_log = None
        self.ext_levels = ['none', 'errors', 'all']
        self.ext_log_file = None
        self.jmx = None
        self.user_args = None
        self.jmeter_path = None
        self.jmeter_ver = None
        self.jmeter_log = None
        self.start_time = time.time()
        self.jmeter_buffer_size = None

    @staticmethod
    def get_key():
        return __file__

    def get_available_options(self):
        return [
            "jmx", "args", "jmeter_path", "buffer_size", "buffered_seconds",
            "exclude_markers"
        ]

    def configure(self):
        self.original_jmx = self.get_option("jmx")
        self.core.add_artifact_file(self.original_jmx, True)
        self.jtl_file = self.core.mkstemp('.jtl', 'jmeter_')
        self.core.add_artifact_file(self.jtl_file)
        self.user_args = self.get_option("args", '')
        self.jmeter_path = self.get_option('jmeter_path', 'jmeter')
        self.jmeter_log = self.core.mkstemp('.log', 'jmeter_')
        self.jmeter_ver = float(self.get_option('jmeter_ver', '3.0'))
        self.ext_log = self.get_option(
            'extended_log', self.get_option('ext_log', 'none'))
        if self.ext_log not in self.ext_levels:
            self.ext_log = 'none'
        if self.ext_log != 'none':
            self.ext_log_file = self.core.mkstemp('.jtl', 'jmeter_ext_')
            self.core.add_artifact_file(self.ext_log_file)
        self.jmeter_buffer_size = int(
            self.get_option(
                'buffer_size', self.get_option('buffered_seconds', '3')))
        self.core.add_artifact_file(self.jmeter_log, True)
        self.exclude_markers = set(
            filter((lambda marker: marker != ''),
                   self.get_option('exclude_markers', []).split(' ')))
        self.jmx = self.__add_jmeter_components(
            self.original_jmx, self.jtl_file, self._get_variables())
        self.core.add_artifact_file(self.jmx)

        jmeter_stderr_file = self.core.mkstemp(".log", "jmeter_stdout_stderr_")
        self.core.add_artifact_file(jmeter_stderr_file)
        self.jmeter_stderr = open(jmeter_stderr_file, 'w')

    def prepare_test(self):
        self.args = [
            self.jmeter_path, "-n", "-t", self.jmx, '-j', self.jmeter_log,
            '-Jjmeter.save.saveservice.default_delimiter=\\t',
            '-Jjmeter.save.saveservice.connect_time=true'
        ]
        self.args += splitstring(self.user_args)

        aggregator = None
        try:
            aggregator = self.core.get_plugin_of_type(AggregatorPlugin)
        except Exception as ex:
            logger.warning("No aggregator found: %s", ex)

        if aggregator:
            aggregator.reader = JMeterReader(self.jtl_file)
            aggregator.stats_reader = aggregator.reader.stats_reader

        try:
            console = self.core.get_plugin_of_type(ConsolePlugin)
        except Exception as ex:
            logger.debug("Console not found: %s", ex)
            console = None

        if console:
            widget = JMeterInfoWidget(self)
            console.add_info_widget(widget)
            if aggregator:
                aggregator.add_result_listener(widget)

    def start_test(self):
        logger.info(
            "Starting %s with arguments: %s", self.jmeter_path, self.args)
        try:
            self.jmeter_process = subprocess.Popen(
                self.args,
                executable=self.jmeter_path,
                preexec_fn=os.setsid,
                close_fds=True,
                stdout=self.jmeter_stderr,
                stderr=self.jmeter_stderr)
        except OSError:
            logger.debug(
                "Unable to start JMeter process. Args: %s, Executable: %s",
                self.args,
                self.jmeter_path,
                exc_info=True)
            raise RuntimeError(
                "Unable to access to JMeter executable file or it does not exist: %s"
                % self.jmeter_path)
        self.start_time = time.time()

    def is_test_finished(self):
        retcode = self.jmeter_process.poll()
        aggregator = self.core.get_plugin_of_type(AggregatorPlugin)
        if not aggregator.reader.jmeter_finished and retcode is not None:
            logger.info(
                "JMeter process finished with exit code: %s, waiting for aggregator",
                retcode)
            self.retries = 0
            aggregator.reader.jmeter_finished = True
            return -1
        elif aggregator.reader.jmeter_finished is True:
            if aggregator.reader.agg_finished:
                return retcode
            else:
                logger.info("Waiting for aggregator to finish")
                return -1
        else:
            return -1

    def end_test(self, retcode):
        if self.jmeter_process:
            logger.info(
                "Terminating jmeter process group with PID %s",
                self.jmeter_process.pid)
            try:
                os.killpg(self.jmeter_process.pid, signal.SIGTERM)
            except OSError as exc:
                logger.debug("Seems JMeter exited itself: %s", exc)
                # Utils.log_stdout_stderr(logger, self.jmeter_process.stdout, self.jmeter_process.stderr, "jmeter")
        if self.jmeter_stderr:
            self.jmeter_stderr.close()
        self.core.add_artifact_file(self.jmeter_log)
        return retcode

    def __add_jmeter_components(self, jmx, jtl, variables):
        """ Genius idea by Alexey Lavrenyuk """
        logger.debug("Original JMX: %s", os.path.realpath(jmx))
        with open(jmx, 'r') as src_jmx:
            source_lines = src_jmx.readlines()

        try:
            closing = source_lines.pop(-1)
            closing = source_lines.pop(-1) + closing
            closing = source_lines.pop(-1) + closing
            logger.debug("Closing statement: %s", closing)
        except Exception as exc:
            raise RuntimeError("Failed to find the end of JMX XML: %s" % exc)

        udv_tpl = resource_string(__name__, 'config/jmeter_var_template.xml')
        udv_set = []
        for var_name, var_value in variables.iteritems():
            udv_set.append(udv_tpl % (var_name, var_name, var_value))
        udv = "\n".join(udv_set)

        if self.jmeter_ver >= 2.13:
            save_connect = '<connectTime>true</connectTime>'
        else:
            save_connect = ''

        if self.ext_log in ['errors', 'all']:
            level_map = {'errors': 'true', 'all': 'false'}
            tpl_resource = 'jmeter_writer_ext.xml'
            tpl_args = {
                'jtl': self.jtl_file,
                'udv': udv,
                'ext_log': self.ext_log_file,
                'ext_level': level_map[self.ext_log],
                'save_connect': save_connect
            }
        else:
            tpl_resource = 'jmeter_writer.xml'
            tpl_args = {
                'jtl': self.jtl_file,
                'udv': udv,
                'save_connect': save_connect
            }

        tpl = resource_string(__name__, 'config/' + tpl_resource)

        try:
            new_jmx = self.core.mkstemp(
                '.jmx', 'modified_', os.path.dirname(os.path.realpath(jmx)))
        except OSError as exc:
            logger.debug("Can't create modified jmx near original: %s", exc)
            new_jmx = self.core.mkstemp('.jmx', 'modified_')
        logger.debug("Modified JMX: %s", new_jmx)
        with open(new_jmx, "wb") as fh:
            fh.write(''.join(source_lines))
            fh.write(tpl % tpl_args)
            fh.write(closing)
        return new_jmx

    def _get_variables(self):
        res = {}
        for option in self.core.config.get_options(self.SECTION):
            if option[0] not in self.get_available_options():
                res[option[0]] = option[1]
        logging.debug("Variables: %s", res)
        return res


class JMeterInfoWidget(AbstractInfoWidget, AggregateResultListener):
    """ Right panel widget with JMeter test info """

    def __init__(self, jmeter):
        AbstractInfoWidget.__init__(self)
        self.krutilka = ConsoleScreen.krutilka()
        self.jmeter = jmeter
        self.active_threads = 0
        self.RPS = 0

    def get_index(self):
        return 0

    def on_aggregated_data(self, data, stats):
        self.active_threads = stats['metrics']['instances']
        self.RPS = data['overall']['interval_real']['len']

    def render(self, screen):
        jmeter = " JMeter Test %s" % self.krutilka.next()
        space = screen.right_panel_width - len(jmeter) - 1
        left_spaces = space / 2
        right_spaces = space / 2

        dur_seconds = int(time.time()) - int(self.jmeter.start_time)
        duration = str(datetime.timedelta(seconds=dur_seconds))

        template = screen.markup.BG_MAGENTA + '~' * left_spaces + jmeter + ' '
        template += '~' * right_spaces + screen.markup.RESET + "\n"
        template += "     Test Plan: %s\n"
        template += "      Duration: %s\n"
        template += "Active Threads: %s\n"
        template += "   Responses/s: %s"
        data = (
            os.path.basename(self.jmeter.original_jmx), duration,
            self.active_threads, self.RPS)

        return template % data



"""
Run maven test as load test
"""
from __future__ import division, absolute_import

import logging
import subprocess
import time

from builtins import super
from ...common.resource import manager as resource_manager
from ...common.interfaces import AbstractPlugin, GeneratorPlugin

from .console import MavenInfoWidget
from .reader import MavenReader, MavenStatsReader
from ..Console import Plugin as ConsolePlugin

logger = logging.getLogger(__name__)


class Plugin(AbstractPlugin, GeneratorPlugin):

    SECTION = "maven"

    def __init__(self, core):
        # FIXME python version 2.7 does not support this syntax. super() should
        # have arguments in Python 2
        super().__init__(core)
        self.maven_cmd = "mvn"
        self.process = None
        self.process_stderr = None
        self.process_start_time = None

    @staticmethod
    def get_key():
        return __file__

    def get_available_options(self):
        opts = ["pom", "testcase", "mvn_args"]
        return opts

    def configure(self):
        # plugin part
        self.pom = resource_manager.resource_filename(
            self.get_option("pom", "pom.xml"))
        self.testcase = self.get_option("testcase", "")
        self.maven_args = self.get_option("mvn_args", '').split()

    def prepare_test(self):
        aggregator = self.core.job.aggregator_plugin

        if aggregator:
            aggregator.reader = MavenReader()
            aggregator.stats_reader = MavenStatsReader()

        try:
            console = self.core.get_plugin_of_type(ConsolePlugin)
        except KeyError as ex:
            logger.debug("Console not found: %s", ex)
            console = None

        if console:
            widget = MavenInfoWidget(self)
            console.add_info_widget(widget)
            if aggregator:
                aggregator.add_result_listener(widget)

    def start_test(self):
        args = [self.maven_cmd, "test", "-Dtest=%s" % self.testcase
                ] + self.maven_args + ["-f", self.pom]
        logger.info("Starting: %s", args)
        self.process_start_time = time.time()
        process_stderr_file = self.core.mkstemp(".log", "maven_")
        self.core.add_artifact_file(process_stderr_file)
        self.process_stderr = open(process_stderr_file, 'w')
        self.process = subprocess.Popen(
            args,
            stderr=self.process_stderr,
            stdout=self.process_stderr,
            close_fds=True)

    def is_test_finished(self):
        retcode = self.process.poll()
        if retcode is not None:
            logger.info("Subprocess done its work with exit code: %s", retcode)
            return abs(retcode)
        else:
            return -1

    def end_test(self, retcode):
        if self.process and self.process.poll() is None:
            logger.warn(
                "Terminating worker process with PID %s", self.process.pid)
            self.process.terminate()
            if self.process_stderr:
                self.process_stderr.close()
        else:
            logger.debug("Seems subprocess finished OK")
        return retcode



"""
Phantom phout format reader. Read chunks from phout and produce data frames
"""
import pandas as pd
import numpy as np
import logging
import json
import time
import datetime
import itertools as itt
from StringIO import StringIO

logger = logging.getLogger(__name__)

phout_columns = [
    'send_ts', 'tag', 'interval_real', 'connect_time', 'send_time', 'latency',
    'receive_time', 'interval_event', 'size_out', 'size_in', 'net_code',
    'proto_code'
]

dtypes = {
    'time': np.float64,
    'tag': np.str,
    'interval_real': np.int64,
    'connect_time': np.int64,
    'send_time': np.int64,
    'latency': np.int64,
    'receive_time': np.int64,
    'interval_event': np.int64,
    'size_out': np.int64,
    'size_in': np.int64,
    'net_code': np.int64,
    'proto_code': np.int64,
}


def string_to_df(data):
    start_time = time.time()
    chunk = pd.read_csv(
        StringIO(data), sep='\t', names=phout_columns, dtype=dtypes)

    chunk['receive_ts'] = chunk.send_ts + chunk.interval_real / 1e6
    chunk['receive_sec'] = chunk.receive_ts.astype(np.int64)
    # TODO: consider configuration for the following:
    chunk['tag'] = chunk.tag.str.rsplit('#', 1, expand=True)[0]
    chunk.set_index(['receive_sec'], inplace=True)

    logger.debug("Chunk decode time: %.2fms", (time.time() - start_time) * 1000)
    return chunk


class PhantomReader(object):
    def __init__(self, filename, cache_size=1024 * 1024 * 50):
        self.buffer = ""
        self.phout = open(filename, 'r')
        self.closed = False
        self.cache_size = cache_size

    def _read_phout_chunk(self):
        data = self.phout.read(self.cache_size)
        if data:
            parts = data.rsplit('\n', 1)
            if len(parts) > 1:
                ready_chunk = self.buffer + parts[0] + '\n'
                self.buffer = parts[1]
                return string_to_df(ready_chunk)
            else:
                self.buffer += parts[0]
        else:
            self.buffer += self.phout.readline()
        return None

    def __iter__(self):
        while not self.closed:
            yield self._read_phout_chunk()
        yield self._read_phout_chunk()
        self.phout.close()

    def close(self):
        self.closed = True


class PhantomStatsReader(object):
    def __init__(self, filename, phantom_info):
        self.phantom_info = phantom_info
        self.buffer = ""
        self.stat_buffer = ""
        self.stat_filename = filename
        self.closed = False
        self.start_time = 0

    def _decode_stat_data(self, chunk):
        """
        Return all items found in this chunk
        """
        for date_str, statistics in chunk.iteritems():
            date_obj = datetime.datetime.strptime(
                date_str.split(".")[0], '%Y-%m-%d %H:%M:%S')
            chunk_date = int(time.mktime(date_obj.timetuple()))
            instances = 0
            for benchmark_name, benchmark in statistics.iteritems():
                if not benchmark_name.startswith("benchmark_io"):
                    continue
                for method, meth_obj in benchmark.iteritems():
                    if "mmtasks" in meth_obj:
                        instances += meth_obj["mmtasks"][2]

            offset = chunk_date - 1 - self.start_time
            reqps = 0
            if offset >= 0 and offset < len(self.phantom_info.steps):
                reqps = self.phantom_info.steps[offset][0]
            yield {
                'ts': chunk_date - 1,
                'metrics': {
                    'instances': instances,
                    'reqps': reqps
                }
            }

    def _read_stat_data(self, stat_file):
        chunk = stat_file.read(1024 * 1024 * 50)
        if chunk:
            self.stat_buffer += chunk
            parts = self.stat_buffer.rsplit('\n},', 1)
            if len(parts) > 1:
                ready_chunk = parts[0]
                self.stat_buffer = parts[1]
                chunks = [
                    json.loads('{%s}}' % s) for s in ready_chunk.split('\n},')
                ]
                return list(
                    itt.chain(
                        *(self._decode_stat_data(chunk) for chunk in chunks)))
        else:
            self.stat_buffer += stat_file.readline()

    def __iter__(self):
        """
        Union buffer and chunk, split using '\n},',
        return splitted parts
        """
        self.start_time = int(time.time())
        with open(self.stat_filename, 'r') as stat_file:
            while not self.closed:
                yield self._read_stat_data(stat_file)
            yield self._read_stat_data(stat_file)

    def close(self):
        self.closed = True



"""Module provides target monitoring
metrics collector - influxdata's `telegraf` - https://github.com/influxdata/telegraf/
backward compatibility with yandextank's Monitoring module configuration and tools.
"""

import datetime
import fnmatch
import json
import logging
import os
import time
from ...common.resource import manager as resource
from ...common.interfaces import MonitoringDataListener, \
    AbstractPlugin, AbstractInfoWidget
from ...common.util import expand_to_seconds

from ..Autostop import Plugin as AutostopPlugin, AbstractCriterion
from ..Console import Plugin as ConsolePlugin
from ..Telegraf.collector import MonitoringCollector

import sys
if sys.version_info[0] < 3:
    from ConfigParser import NoOptionError
else:
    from configparser import NoOptionError

logger = logging.getLogger(__name__)


class Plugin(AbstractPlugin):
    """  resource mon plugin  """

    SECTION = 'telegraf'  # may be redefined to 'monitoring' sometimes.

    def __init__(self, core):
        super(Plugin, self).__init__(core)
        self.jobno = None
        self.default_target = None
        self.default_config = "{path}/config/monitoring_default_config.xml".format(
            path=os.path.dirname(__file__))
        self.config = None
        self.process = None
        self.monitoring = MonitoringCollector()
        self.die_on_fail = True
        self.data_file = None
        self.mon_saver = None

    @staticmethod
    def get_key():
        return __file__

    def start_test(self):
        if self.monitoring:
            self.monitoring.load_start_time = time.time()
            logger.debug(
                "load_start_time = %s", self.monitoring.load_start_time)

    def get_available_options(self):
        return ["config", "default_target", "ssh_timeout"]

    def __detect_configuration(self):
        """
        we need to be flexible in order to determine which plugin's configuration
        specified and make appropriate configs to metrics collector

        :return: SECTION name or None for defaults
        """
        try:
            is_telegraf = self.core.get_option('telegraf', "config", None)
        except NoOptionError:
            is_telegraf = None
        try:
            is_monitoring = self.core.get_option('monitoring', "config", None)
        except NoOptionError:
            is_monitoring = None

        if is_telegraf and is_monitoring:
            raise ValueError(
                'Both telegraf and monitoring configs specified. '
                'Clean up your config and delete one of them')
        if is_telegraf and not is_monitoring:
            return 'telegraf'
        if not is_telegraf and is_monitoring:
            return 'monitoring'
        if not is_telegraf and not is_monitoring:
            # defaults target logic
            try:
                is_telegraf_dt = self.core.get_option(
                    'telegraf', "default_target", None)
            except NoOptionError:
                is_telegraf_dt = None
            try:
                is_monitoring_dt = self.core.get_option(
                    'monitoring', "default_target", None)
            except:
                is_monitoring_dt = None
            if is_telegraf_dt and is_monitoring_dt:
                raise ValueError(
                    'Both telegraf and monitoring default targets specified. '
                    'Clean up your config and delete one of them')
            if is_telegraf_dt and not is_monitoring_dt:
                return
            if not is_telegraf_dt and is_monitoring_dt:
                self.core.set_option(
                    "telegraf", "default_target", is_monitoring_dt)
            if not is_telegraf_dt and not is_monitoring_dt:
                return

    def configure(self):
        self.detected_conf = self.__detect_configuration()
        if self.detected_conf:
            logging.info(
                'Detected monitoring configuration: %s', self.detected_conf)
            self.SECTION = self.detected_conf
        self.config = self.get_option("config", "auto").strip()
        self.default_target = self.get_option("default_target", "localhost")
        if self.config.lower() == "none":
            self.monitoring = None
            self.die_on_fail = False
            return

        # FIXME [legacy] backward compatibility with Monitoring module
        # configuration below.
        self.monitoring.ssh_timeout = expand_to_seconds(
            self.get_option("ssh_timeout", "5s"))

        # FIXME [legacy] handle raw XML config in .ini file
        if self.config[0] == "<":
            config_contents = self.config
        # handle http/https url or file path
        else:
            if self.config.lower() == "auto":
                self.die_on_fail = False
                with open(
                        resource.resource_filename(self.default_config),
                        'rb') as def_config:
                    config_contents = def_config.read()
            else:
                with open(resource.resource_filename(self.config),
                          'rb') as config:
                    config_contents = config.read()

        # dump config contents into a file
        xmlfile = self.core.mkstemp(".xml", "monitoring_")
        self.core.add_artifact_file(xmlfile)
        with open(
                xmlfile, "wb"
        ) as f:  # output file should be in binary mode to support py3
            f.write(config_contents)
        self.config = xmlfile

        try:
            autostop = self.core.get_plugin_of_type(AutostopPlugin)
            autostop.add_criterion_class(MetricHigherCriterion)
            autostop.add_criterion_class(MetricLowerCriterion)
        except KeyError:
            logger.debug(
                "No autostop plugin found, not adding instances criterion")

    def prepare_test(self):
        if not self.config or self.config.lower() == 'none':
            return

        if "Phantom" in self.core.job.generator_plugin.__module__:
            phantom = self.core.job.generator_plugin
            if phantom.phout_import_mode:
                logger.info("Phout import mode, disabling monitoring")
                self.config = None
                self.monitoring = None

            info = phantom.get_info()
            if info:
                self.default_target = info.address
                logger.debug(
                    "Changed monitoring target to %s", self.default_target)

        self.monitoring.config = self.config
        if self.default_target:
            self.monitoring.default_target = self.default_target

        # FIXME json report already save this artifact, fix pls
        self.data_file = self.core.mkstemp(".data", "monitoring_overall_")
        self.mon_saver = SaveMonToFile(self.data_file)
        self.monitoring.add_listener(self.mon_saver)
        self.core.add_artifact_file(self.data_file)

        try:
            console = self.core.get_plugin_of_type(ConsolePlugin)
        except Exception as ex:
            logger.debug("Console not found: %s", ex)
            console = None
        if console:
            widget = MonitoringWidget(self)
            console.add_info_widget(widget)
            self.monitoring.add_listener(widget)

        try:
            self.monitoring.prepare()
            self.monitoring.start()
            count = 0
            while not self.monitoring.first_data_received and count < 15 * 5:
                time.sleep(0.2)
                self.monitoring.poll()
                count += 1
        except:
            logger.error("Could not start monitoring", exc_info=True)
            if self.die_on_fail:
                raise
            else:
                self.monitoring = None

    def is_test_finished(self):
        if self.monitoring:
            data_len = self.monitoring.poll()
            logger.debug("Monitoring got %s lines", data_len)
        return -1

    def end_test(self, retcode):
        logger.info("Finishing monitoring")
        if self.monitoring:
            self.monitoring.stop()
            for log in self.monitoring.artifact_files:
                self.core.add_artifact_file(log)

            while self.monitoring.send_data:
                logger.info("Sending monitoring data rests...")
                self.monitoring.send_collected_data()
        if self.mon_saver:
            self.mon_saver.close()
        return retcode


class SaveMonToFile(MonitoringDataListener):
    """
    Default listener - saves data to file
    """

    def __init__(self, out_file):
        MonitoringDataListener.__init__(self)
        if out_file:
            self.store = open(out_file, 'w')

    def monitoring_data(self, data):
        self.store.write(json.dumps(data))
        self.store.write('\n')
        self.store.flush()

    def close(self):
        """ close open files """
        logger.debug("Closing monitoring file")
        if self.store:
            self.store.close()


class MonitoringWidget(AbstractInfoWidget, MonitoringDataListener):
    """
    Screen widget
    """

    def __init__(self, owner):
        AbstractInfoWidget.__init__(self)
        self.owner = owner
        self.data = {}
        self.sign = {}
        self.time = {}
        self.max_metric_len = 0

    def get_index(self):
        return 50

    def __handle_data_items(self, host, data):
        """ store metric in data tree and calc offset signs

        sign < 0 is CYAN, means metric value is lower then previous,
        sign > 1 is YELLOW, means metric value is higher then prevoius,
        sign == 0 is WHITE, means initial or equal metric value
        """
        for metric, value in data.iteritems():
            if value == '':
                self.sign[host][metric] = -1
                self.data[host][metric] = value
            else:
                if not self.data[host].get(metric, None):
                    self.sign[host][metric] = 1
                elif float(value) > float(self.data[host][metric]):
                    self.sign[host][metric] = 1
                elif float(value) < float(self.data[host][metric]):
                    self.sign[host][metric] = -1
                else:
                    self.sign[host][metric] = 0
                self.data[host][metric] = "%.2f" % float(value)

    def monitoring_data(self, block):
        # block sample :
        # [{'timestamp': 1480536634,
        #   'data': {
        #     'some.hostname.tld': {
        #       'comment': '',
        #       'metrics': {
        #         'custom:diskio_reads': 0,
        #         'Net_send': 9922,
        #         'CPU_steal': 0,
        #         'Net_recv': 8489
        #       }
        #     }
        #   },
        #   ...
        # }]
        for chunk in block:
            host = chunk['data'].keys()[0]
            self.time[host] = chunk['timestamp']
            # if initial call, we create dicts w/ data and `signs`
            # `signs` used later to paint metrics w/ different colors
            if not self.data.get(host, None):
                self.data[host] = {}
                self.sign[host] = {}
                for key, value in chunk['data'][host]['metrics'].iteritems():
                    self.sign[host][key] = 0
                    self.data[host][key] = value
            else:
                self.__handle_data_items(host, chunk['data'][host]['metrics'])

    def render(self, screen):
        if not self.owner.monitoring:
            return "Monitoring is " + screen.markup.RED + "offline" + screen.markup.RESET
        else:
            res = "Monitoring is " + screen.markup.GREEN + \
                "online" + screen.markup.RESET + ":\n"
            for hostname, metrics in self.data.items():
                tm_stamp = datetime.datetime.fromtimestamp(
                    float(self.time[hostname])).strftime('%H:%M:%S')
                res += (
                    "   " + screen.markup.CYAN + "%s" + screen.markup.RESET +
                    " at %s:\n") % (hostname, tm_stamp)
                for metric, value in sorted(metrics.iteritems()):
                    if self.sign[hostname][metric] > 0:
                        value = screen.markup.YELLOW + value + screen.markup.RESET
                    elif self.sign[hostname][metric] < 0:
                        value = screen.markup.CYAN + value + screen.markup.RESET
                    res += "      %s%s: %s\n" % (
                        ' ' * (self.max_metric_len - len(metric)),
                        metric.replace('custom:', '').replace('_', ' '), value)

            return res.strip()


class AbstractMetricCriterion(AbstractCriterion, MonitoringDataListener):
    """ Parent class for metric criterion """

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        try:
            self.mon = autostop.core.get_plugin_of_type(Plugin)
            if self.mon.monitoring:
                self.mon.monitoring.add_listener(self)
        except KeyError:
            logger.warning("No monitoring module, mon autostop disabled")
        self.triggered = False
        self.autostop = autostop

        self.host = param_str.split(',')[0].strip()
        self.metric = param_str.split(',')[1].strip()
        self.value_limit = float(param_str.split(',')[2])
        self.seconds_limit = expand_to_seconds(param_str.split(',')[3])
        self.last_second = None
        self.seconds_count = 0

    def monitoring_data(self, block):
        if self.triggered:
            return

        for chunk in block:
            host = chunk['data'].keys()[0]
            data = chunk['data'][host]['metrics']

            if not fnmatch.fnmatch(host, self.host):
                continue

            # some magic, converting custom metric names into names that was in
            # config
            for metric_name in data.keys():
                if metric_name.startswith('custom:'):
                    config_metric_name = metric_name.replace('custom:', '')
                    data[config_metric_name] = data.pop(metric_name)

            if self.metric not in data.keys() or not data[self.metric]:
                data[self.metric] = 0
            logger.debug(
                "Compare %s %s/%s=%s to %s",
                self.get_type_string(), host, self.metric, data[self.metric],
                self.value_limit)
            if self.comparison_fn(float(data[self.metric]), self.value_limit):
                if not self.seconds_count:
                    self.cause_second = self.last_second

                logger.debug(self.explain())

                self.seconds_count += 1
                if self.seconds_count >= self.seconds_limit:
                    logger.debug("Triggering autostop")
                    self.triggered = True
                    return
            else:
                self.seconds_count = 0

    def notify(self, data, stat):
        if self.seconds_count:
            self.autostop.add_counting(self)

        self.last_second = (data, stat)
        return self.triggered

    def comparison_fn(self, arg1, arg2):
        """ comparison function """
        raise NotImplementedError()


class MetricHigherCriterion(AbstractMetricCriterion):
    """ trigger if metric is higher than limit """

    def __init__(self, autostop, param_str):
        AbstractMetricCriterion.__init__(self, autostop, param_str)

    def get_rc(self):
        return 31

    @staticmethod
    def get_type_string():
        return 'metric_higher'

    def explain(self):
        items = (self.host, self.metric, self.value_limit, self.seconds_count)
        return "%s/%s metric value is higher than %s for %s seconds" % items

    def widget_explain(self):
        items = (
            self.host, self.metric, self.value_limit, self.seconds_count,
            self.seconds_limit)
        return "%s/%s > %s for %s/%ss" % items, float(
            self.seconds_count) / self.seconds_limit

    def comparison_fn(self, arg1, arg2):
        return arg1 > arg2


class MetricLowerCriterion(AbstractMetricCriterion):
    """ trigger if metric is lower than limit """

    def __init__(self, autostop, param_str):
        AbstractMetricCriterion.__init__(self, autostop, param_str)

    def get_rc(self):
        return 32

    @staticmethod
    def get_type_string():
        return 'metric_lower'

    def explain(self):
        items = (self.host, self.metric, self.value_limit, self.seconds_count)
        return "%s/%s metric value is lower than %s for %s seconds" % items

    def widget_explain(self):
        items = (
            self.host, self.metric, self.value_limit, self.seconds_count,
            self.seconds_limit)
        return "%s/%s < %s for %s/%ss" % items, float(
            self.seconds_count) / self.seconds_limit

    def comparison_fn(self, arg1, arg2):
        return arg1 < arg2



import copy
import json
import logging
import re
import time

from yandextank.common.util import expand_to_seconds, expand_to_milliseconds
from ...common.interfaces import AbstractCriterion

logger = logging.getLogger(__name__)


class AvgTimeCriterion(AbstractCriterion):
    """ average response time criterion """

    @staticmethod
    def get_type_string():
        return 'time'

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        self.seconds_count = 0
        self.rt_limit = expand_to_milliseconds(param_str.split(',')[0])
        self.seconds_limit = expand_to_seconds(param_str.split(',')[1])
        self.autostop = autostop

    def notify(self, data, stat):
        if (
                data["overall"]["interval_real"]["total"] / 1000.0 /
                data["overall"]["interval_real"]["len"]) > self.rt_limit:
            if not self.seconds_count:
                self.cause_second = (data, stat)

            logger.debug(self.explain())

            self.seconds_count += 1
            self.autostop.add_counting(self)
            if self.seconds_count >= self.seconds_limit:
                return True
        else:
            self.seconds_count = 0

        return False

    def get_rc(self):
        return self.RC_TIME

    def explain(self):
        explanation = (
            "Average response time higher"
            " than %sms for %ss, since %s" %
            (self.rt_limit, self.seconds_count, self.cause_second[0]["ts"]))
        return explanation

    def widget_explain(self):
        items = (self.rt_limit, self.seconds_count, self.seconds_limit)
        return "Avg Time >%sms for %s/%ss" % items, float(
            self.seconds_count) / self.seconds_limit


class HTTPCodesCriterion(AbstractCriterion):
    """ HTTP codes criterion """

    @staticmethod
    def get_type_string():
        return 'http'

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        self.seconds_count = 0
        self.codes_mask = param_str.split(',')[0].lower()
        self.codes_regex = re.compile(self.codes_mask.replace("x", '.'))
        self.autostop = autostop

        level_str = param_str.split(',')[1].strip()
        if level_str[-1:] == '%':
            self.level = float(level_str[:-1]) / 100
            self.is_relative = True
        else:
            self.level = int(level_str)
            self.is_relative = False
        self.seconds_limit = expand_to_seconds(param_str.split(',')[2])

    def notify(self, data, stat):
        matched_responses = self.count_matched_codes(
            self.codes_regex, data["overall"]["proto_code"]["count"])
        if self.is_relative:
            if data["overall"]["interval_real"]["len"]:
                matched_responses = float(matched_responses) / data["overall"][
                    "interval_real"]["len"]
            else:
                matched_responses = 0
        logger.debug(
            "HTTP codes matching mask %s: %s/%s", self.codes_mask,
            matched_responses, self.level)

        if matched_responses >= self.level:
            if not self.seconds_count:
                self.cause_second = (data, stat)

            logger.debug(self.explain())

            self.seconds_count += 1
            self.autostop.add_counting(self)
            if self.seconds_count >= self.seconds_limit:
                return True
        else:
            self.seconds_count = 0

        return False

    def get_rc(self):
        return self.RC_HTTP

    def get_level_str(self):
        """ format level str """
        if self.is_relative:
            level_str = str(100 * self.level) + "%"
        else:
            level_str = self.level
        return level_str

    def explain(self):
        items = (
            self.codes_mask, self.get_level_str(), self.seconds_count,
            self.cause_second[0].get('ts'))
        return "%s codes count higher than %s for %ss, since %s" % items

    def widget_explain(self):
        items = (
            self.codes_mask, self.get_level_str(), self.seconds_count,
            self.seconds_limit)
        return "HTTP %s>%s for %s/%ss" % items, float(
            self.seconds_count) / self.seconds_limit


class NetCodesCriterion(AbstractCriterion):
    """ Net codes criterion """

    @staticmethod
    def get_type_string():
        return 'net'

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        self.seconds_count = 0
        self.codes_mask = param_str.split(',')[0].lower()
        self.codes_regex = re.compile(self.codes_mask.replace("x", '.'))
        self.autostop = autostop

        level_str = param_str.split(',')[1].strip()
        if level_str[-1:] == '%':
            self.level = float(level_str[:-1]) / 100
            self.is_relative = True
        else:
            self.level = int(level_str)
            self.is_relative = False
        self.seconds_limit = expand_to_seconds(param_str.split(',')[2])

    def notify(self, data, stat):
        codes = copy.deepcopy(data["overall"]["net_code"]["count"])
        if '0' in codes.keys():
            codes.pop('0')
        matched_responses = self.count_matched_codes(self.codes_regex, codes)
        if self.is_relative:
            if data["overall"]["interval_real"]["len"]:
                matched_responses = float(matched_responses) / data["overall"][
                    "interval_real"]["len"]
            else:
                matched_responses = 0
        logger.debug(
            "Net codes matching mask %s: %s/%s", self.codes_mask,
            matched_responses, self.level)

        if matched_responses >= self.level:
            if not self.seconds_count:
                self.cause_second = (data, stat)

            logger.debug(self.explain())

            self.seconds_count += 1
            self.autostop.add_counting(self)
            if self.seconds_count >= self.seconds_limit:
                return True
        else:
            self.seconds_count = 0

        return False

    def get_rc(self):
        return self.RC_NET

    def get_level_str(self):
        """ format level str """
        if self.is_relative:
            level_str = str(100 * self.level) + "%"
        else:
            level_str = self.level
        return level_str

    def explain(self):
        items = (
            self.codes_mask, self.get_level_str(), self.seconds_count,
            self.cause_second[0].get("ts"))
        return "%s net codes count higher than %s for %ss, since %s" % items

    def widget_explain(self):
        items = (
            self.codes_mask, self.get_level_str(), self.seconds_count,
            self.seconds_limit)
        return "Net %s>%s for %s/%ss" % items, float(
            self.seconds_count) / self.seconds_limit


class QuantileCriterion(AbstractCriterion):
    """ quantile criterion """

    @staticmethod
    def get_type_string():
        return 'quantile'

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        self.seconds_count = 0
        self.quantile = float(param_str.split(',')[0])
        self.rt_limit = expand_to_milliseconds(param_str.split(',')[1])
        self.seconds_limit = expand_to_seconds(param_str.split(',')[2])
        self.autostop = autostop

    def notify(self, data, stat):
        quantiles = dict(
            zip(
                data["overall"]["interval_real"]["q"]["q"], data["overall"][
                    "interval_real"]["q"]["value"]))
        if self.quantile not in quantiles.keys():
            logger.warning("No quantile %s in %s", self.quantile, quantiles)
        if self.quantile in quantiles.keys() \
                and quantiles[self.quantile] / 1000.0 > self.rt_limit:
            if not self.seconds_count:
                self.cause_second = (data, stat)

            logger.debug(self.explain())

            self.seconds_count += 1
            self.autostop.add_counting(self)
            if self.seconds_count >= self.seconds_limit:
                return True
        else:
            self.seconds_count = 0

        return False

    def get_rc(self):
        return self.RC_TIME

    def explain(self):
        items = (
            self.quantile, self.rt_limit, self.seconds_count,
            self.cause_second[0].get("ts"))
        return "Percentile %s higher than %sms for %ss, since %s" % items

    def widget_explain(self):
        items = (
            self.quantile, self.rt_limit, self.seconds_count,
            self.seconds_limit)
        return "%s%% >%sms for %s/%ss" % items, float(
            self.seconds_count) / self.seconds_limit


class SteadyCumulativeQuantilesCriterion(AbstractCriterion):
    """ quantile criterion """

    @staticmethod
    def get_type_string():
        return 'steady_cumulative'

    def __init__(self, autostop, param_str):
        raise NotImplementedError
        AbstractCriterion.__init__(self)
        self.seconds_count = 0
        self.quantile_hash = ""
        self.seconds_limit = expand_to_seconds(param_str.split(',')[0])
        self.autostop = autostop

    def notify(self, data, stat):
        quantiles = dict(
            zip(data["overall"]["q"]["q"], data["overall"]["q"]["values"]))
        quantile_hash = json.dumps(quantiles)
        logging.debug("Cumulative quantiles hash: %s", quantile_hash)
        if self.quantile_hash == quantile_hash:
            if not self.seconds_count:
                self.cause_second = (data, stat)

            logger.debug(self.explain())

            self.seconds_count += 1
            self.autostop.add_counting(self)
            if self.seconds_count >= self.seconds_limit:
                return True
        else:
            self.seconds_count = 0

        self.quantile_hash = quantile_hash
        return False

    def get_rc(self):
        return self.RC_STEADY

    def explain(self):
        items = (self.seconds_count, self.cause_second[0]["ts"])
        return "Cumulative percentiles are steady for %ss, since %s" % items

    def widget_explain(self):
        items = (self.seconds_count, self.seconds_limit)
        return "Steady for %s/%ss" % items, float(
            self.seconds_count) / self.seconds_limit


class TimeLimitCriterion(AbstractCriterion):
    """ time limit criterion """

    @staticmethod
    def get_type_string():
        return 'limit'

    def __init__(self, autostop, param_str):
        AbstractCriterion.__init__(self)
        self.start_time = time.time()
        self.end_time = time.time()
        self.time_limit = expand_to_seconds(param_str)

    def notify(self, data, stat):
        self.end_time = time.time()
        return (self.end_time - self.start_time) > self.time_limit

    def get_rc(self):
        return self.RC_TIME

    def explain(self):
        return "Test time elapsed. Limit: %ss, actual time: %ss" % (
            self.time_limit, self.end_time - self.start_time)

    def widget_explain(self):
        return "Time limit: %ss, actual time: %ss" % (
            self.time_limit, self.end_time - self.start_time)



import datetime
import json
import time
import urllib
import requests
import logging

requests.packages.urllib3.disable_warnings()
logger = logging.getLogger(__name__)  # pylint: disable=C0103


class OverloadClient(object):
    def __init__(self):
        self.address = None
        self.token = None
        self.upload_token = ''
        self.api_token = ""
        self.api_timeout = None
        self.session = requests.Session()
        self.session.verify = False
        self.session.headers.update({"User-Agent": "tank"})

    def set_api_address(self, addr):
        self.address = addr

    def set_api_timeout(self, timeout):
        self.api_timeout = float(timeout)

    def set_api_token(self, api_token):
        self.api_token = api_token

    def get_raw(self, addr):
        if not self.address:
            raise ValueError("Can't request unknown address")

        addr = self.address + addr
        logger.debug("Making request to: %s", addr)
        req = requests.Request('GET', addr)
        prepared = self.session.prepare_request(req)
        resp = self.session.send(prepared, timeout=self.api_timeout)
        resp.raise_for_status()
        resp_data = resp.content.strip()
        logger.debug("Raw response: %s", resp_data)
        return resp_data

    def get(self, addr):
        resp = self.get_raw(addr)
        response = json.loads(resp)
        logger.debug("Response: %s", response)
        return response

    def post_raw(self, addr, txt_data):
        if not self.address:
            raise ValueError("Can't request unknown address")

        addr = self.address + addr
        logger.debug("Making POST request to: %s", addr)
        req = requests.Request("POST", addr, data=txt_data)
        prepared = self.session.prepare_request(req)
        resp = self.session.send(prepared, timeout=self.api_timeout)
        resp.raise_for_status()
        logger.debug("Response: %s", resp.content)
        return resp.content

    def post(self, addr, data):
        addr = self.address + addr
        json_data = json.dumps(data, indent=2)
        logger.debug("Making POST request to: %s\n%s", addr, json_data)
        req = requests.Request("POST", addr, data=json_data)
        prepared = self.session.prepare_request(req)
        resp = self.session.send(prepared, timeout=self.api_timeout)
        resp.raise_for_status()
        logger.debug("Response: %s", resp.content)
        return resp.json()

    def get_task_data(self, task):
        return self.get("api/task/" + task + "/summary.json")

    def new_job(
            self, task, person, tank, target_host, target_port, loadscheme,
            detailed_time, notify_list):
        data = {
            'task': task,
            'person': person,
            'tank': tank,
            'host': target_host,
            'port': target_port,
            'loadscheme': loadscheme,
            'detailed_time': detailed_time,
            'notify': notify_list,
        }

        logger.debug("Job create request: %s", data)
        while True:
            try:
                response = self.post(
                    "api/job/create.json?api_token=" + self.api_token, data)
                self.upload_token = response[0].get('upload_token', '')
                return response[0]['job']
            except requests.exceptions.HTTPError as ex:
                logger.debug("Got error for job create request: %s", ex)
                if ex.response.status_code == 423:
                    logger.warn(
                        "Overload is under maintenance, will retry in 5s...")
                    time.sleep(5)
                else:
                    raise ex

        raise RuntimeError("Unreachable point hit")

    def get_job_summary(self, jobno):
        result = self.get(
            'api/job/' + str(jobno) + "/summary.json?api_token=" +
            self.api_token)
        return result[0]

    def close_job(self, jobno, retcode):
        params = {
            'exitcode': str(retcode),
            'api_token': self.api_token,
        }

        result = self.get(
            'api/job/' + str(jobno) + '/close.json?' + urllib.urlencode(params))
        return result[0]['success']

    def edit_job_metainfo(
            self, jobno, job_name, job_dsc, instances, ammo_path, loop_count,
            version_tested, is_regression, component, tank_type, cmdline,
            is_starred):
        data = {
            'name': job_name,
            'description': job_dsc,
            'instances': str(instances),
            'ammo': ammo_path,
            'loop': loop_count,
            'version': version_tested,
            'regression': str(is_regression),
            'component': component,
            'tank_type': int(tank_type),
            'command_line': cmdline,
            'starred': int(is_starred),
        }

        response = self.post(
            'api/job/' + str(jobno) + "/edit.json?api_token=" + self.api_token,
            data)
        return response

    def set_imbalance_and_dsc(self, jobno, rps, comment):
        data = {}
        if rps:
            data['imbalance'] = rps
        if comment:
            data['description'] = comment.strip()

        response = self.post(
            'api/job/' + str(jobno) + "/set_imbalance.json?api_token=" +
            self.api_token, data)
        return response

    def second_data_to_push_item(self, data, stat, timestamp, overall, case):
        """
        @data: SecondAggregateDataItem
        """
        api_data = {
            'overall': overall,
            'case': case,
            'net_codes': [],
            'http_codes': [],
            'time_intervals': [],
            'trail': {
                'time': str(timestamp),
                'reqps': stat["metrics"]["reqps"],
                'resps': data["interval_real"]["len"],
                'expect': data["interval_real"]["total"] / 1000.0 /
                data["interval_real"]["len"],
                'disper': 0,
                'self_load':
                0,  # TODO abs(round(100 - float(data.selfload), 2)),
                'input': data["size_in"]["total"],
                'output': data["size_out"]["total"],
                'connect_time': data["connect_time"]["total"] / 1000.0 /
                data["connect_time"]["len"],
                'send_time':
                data["send_time"]["total"] / 1000.0 / data["send_time"]["len"],
                'latency':
                data["latency"]["total"] / 1000.0 / data["latency"]["len"],
                'receive_time': data["receive_time"]["total"] / 1000.0 /
                data["receive_time"]["len"],
                'threads': stat["metrics"]["instances"],  # TODO
            }
        }

        for q, value in zip(
                data["interval_real"]["q"]["q"],
                data["interval_real"]["q"]["value"]):
            api_data['trail']['q' + str(q)] = value / 1000.0

        for code, cnt in data["net_code"]["count"].iteritems():
            api_data['net_codes'].append({'code': int(code), 'count': int(cnt)})

        for code, cnt in data["proto_code"]["count"].iteritems():
            api_data['http_codes'].append({
                'code': int(code),
                'count': int(cnt)
            })

        api_data['time_intervals'] = self.convert_hist(
            data["interval_real"]["hist"])
        return api_data

    def convert_hist(self, hist):
        data = hist['data']
        bins = hist['bins']
        return [
            {
                "from": 0,  # deprecated
                "to": b / 1000.0,
                "count": count,
            } for b, count in zip(bins, data)
        ]

    def push_test_data(self, jobno, data_item, stat_item):
        items = []
        uri = 'api/job/{0}/push_data.json?upload_token={1}'.format(
            jobno, self.upload_token)
        ts = datetime.datetime.fromtimestamp(data_item["ts"])
        for case_name, case_data in data_item["tagged"].iteritems():
            if case_name == "":
                case_name = "__EMPTY__"
            if (len(case_name)) > 128:
                raise RuntimeError('tag (case) name is too long: ' + case_name)
            push_item = self.second_data_to_push_item(
                case_data, stat_item, ts, 0, case_name)
            items.append(push_item)
        overall = self.second_data_to_push_item(
            data_item["overall"], stat_item, ts, 1, '')
        items.append(overall)

        while True:
            try:
                res = self.post(uri, items)
                break
            except requests.exceptions.HTTPError as ex:
                if ex.response.status_code == 400:
                    logger.error('Bad request to %s: %s', uri, ex)
                    return 0
                elif ex.response.status_code == 410:
                    logger.info("Test has been stopped by Overload server")
                    return 1
                else:
                    logger.warn(
                        "Unknown HTTP error while sending second data. "
                        "Retry in 10 sec: %s", ex)
                    time.sleep(10)  # FIXME this makes all plugins freeze
            except requests.exceptions.RequestException as ex:
                logger.warn(
                    "Failed to push second data to API,"
                    " retry in 10 sec: %s", ex)
                time.sleep(10)  # FIXME this makes all plugins freeze
            except Exception:  # pylint: disable=W0703
                # something nasty happened, but we don't want to fail here
                logger.exception(
                    "Unknown exception while pushing second data to API")
                return 0
        try:
            success = int(res[0]['success'])
        except Exception:  # pylint: disable=W0703
            logger.warning("Malformed response from API: %s", res)
            success = 0
        return success

    def push_monitoring_data(self, jobno, send_data):
        if send_data:
            addr = "api/monitoring/receiver/push?job_id=%s&upload_token=%s" % (
                jobno, self.upload_token)
            while True:
                try:
                    self.post_raw(addr, send_data)
                    return
                except requests.exceptions.HTTPError as ex:
                    if ex.response.status_code == 400:
                        logger.error('Bad request to %s: %s', addr, ex)
                        break
                    elif ex.response.status_code == 410:
                        logger.info("Test has been stopped by Overload server")
                        return
                    else:
                        logger.warning(
                            'Unknown http code while sending monitoring data,'
                            ' retry in 10s: %s', ex)
                        time.sleep(10)  # FIXME this makes all plugins freeze
                except requests.exceptions.RequestException as ex:
                    logger.warning(
                        'Problems sending monitoring data,'
                        ' retry in 10s: %s', ex)
                    time.sleep(10)  # FIXME this makes all plugins freeze
                except Exception:  # pylint: disable=W0703
                    # something irrecoverable happened
                    logger.exception(
                        "Unknown exception while pushing monitoring data to API")
                    return

    def send_console(self, jobno, console):
        logger.debug(
            "Sending console view [%s]: %s", len(console), console[:64])
        addr = ("api/job/%s/console.txt?api_token=" % jobno) + self.api_token,
        self.post_raw(addr, {"console": console, })

    def send_config_snapshot(self, jobno, config):
        logger.debug("Sending config snapshot")
        addr = ("api/job/%s/configinfo.txt?api_token=" % jobno) + self.api_token
        self.post_raw(addr, {"configinfo": config, })



""" Core module to calculate aggregate data """
import json
import logging

import queue as q
from pkg_resources import resource_string
from ...common.exceptions import PluginImplementationError

from .aggregator import Aggregator, DataPoller
from .chopper import TimeChopper
from ...common.interfaces import AbstractPlugin
from ...common.interfaces import AggregateResultListener
from ...common.util import Drain, Chopper

logger = logging.getLogger(__name__)


class LoggingListener(AggregateResultListener):
    """ Log aggregated results """

    def on_aggregated_data(self, data, stats):
        logger.info("Got aggregated sample:\n%s", json.dumps(data, indent=2))
        logger.info("Stats:\n%s", json.dumps(stats, indent=2))


def get_from_queue(queue):
    data = []
    for _ in range(queue.qsize()):
        try:
            data.append(queue.get_nowait())
        except q.Empty:
            break
    return data


class Plugin(AbstractPlugin):
    """
    Plugin that manages aggregation and stats collection
    """

    SECTION = 'aggregator'

    @staticmethod
    def get_key():
        return __file__

    def __init__(self, core):
        AbstractPlugin.__init__(self, core)
        self.listeners = []  # [LoggingListener()]
        self.reader = None
        self.stats_reader = None
        self.results = q.Queue()
        self.stats = q.Queue()
        self.verbose_histogram = False
        self.data_cache = {}
        self.stat_cache = {}

    def get_available_options(self):
        return ["verbose_histogram"]

    def configure(self):
        self.aggregator_config = json.loads(
            resource_string(__name__, 'config/phout.json').decode('utf8'))
        verbose_histogram_option = self.get_option("verbose_histogram", "0")
        self.verbose_histogram = (
            verbose_histogram_option.lower() == "true") or (
                verbose_histogram_option.lower() == "1")
        if self.verbose_histogram:
            logger.info("using verbose histogram")

    def start_test(self):
        if self.reader and self.stats_reader:
            pipeline = Aggregator(
                TimeChopper(
                    DataPoller(
                        source=self.reader, poll_period=1), cache_size=3),
                self.aggregator_config,
                self.verbose_histogram)
            self.drain = Drain(pipeline, self.results)
            self.drain.start()
            self.stats_drain = Drain(
                Chopper(DataPoller(
                    source=self.stats_reader, poll_period=1)),
                self.stats)
            self.stats_drain.start()
        else:
            raise PluginImplementationError(
                "Generator must pass a Reader and a StatsReader"
                " to Aggregator before starting test")

    def _collect_data(self):
        """
        Collect data, cache it and send to listeners
        """
        data = get_from_queue(self.results)
        stats = get_from_queue(self.stats)
        logger.debug("Data timestamps:\n%s" % [d.get('ts') for d in data])
        logger.debug("Stats timestamps:\n%s" % [d.get('ts') for d in stats])
        for item in data:
            ts = item['ts']
            if ts in self.stat_cache:
                # send items
                data_item = item
                stat_item = self.stat_cache.pop(ts)
                self.__notify_listeners(data_item, stat_item)
            else:
                self.data_cache[ts] = item
        for item in stats:
            ts = item['ts']
            if ts in self.data_cache:
                # send items
                data_item = self.data_cache.pop(ts)
                stat_item = item
                self.__notify_listeners(data_item, stat_item)
            else:
                self.stat_cache[ts] = item

    def is_test_finished(self):
        self._collect_data()
        return -1

    def end_test(self, retcode):
        if self.reader:
            self.reader.close()
        if self.drain:
            self.drain.join()
        if self.stats_reader:
            self.stats_reader.close()
        if self.stats_drain:
            self.stats_drain.join()
        self._collect_data()
        return retcode

    def add_result_listener(self, listener):
        self.listeners.append(listener)

    def __notify_listeners(self, data, stats):
        """ notify all listeners about aggregate data and stats """
        for listener in self.listeners:
            listener.on_aggregated_data(data, stats)



# TODO: make the next two lines unnecessary
# pylint: disable=line-too-long
# pylint: disable=missing-docstring
import StringIO
import copy
import json
import logging
import os
import pwd
import socket
import sys

from .client import OverloadClient
from ..Autostop import Plugin as AutostopPlugin
from ..Console import Plugin as ConsolePlugin
from ..JMeter import Plugin as JMeterPlugin
from ..Monitoring import Plugin as MonitoringPlugin
# from ..Pandora import Plugin as PandoraPlugin
from ..Phantom import Plugin as PhantomPlugin
from ...common.interfaces import AbstractPlugin,\
    MonitoringDataListener, AggregateResultListener, AbstractInfoWidget

logger = logging.getLogger(__name__)  # pylint: disable=C0103


class Plugin(AbstractPlugin, AggregateResultListener, MonitoringDataListener):
    """
    Yandex Overload analytics service client (https://overload.yandex.net)
    """
    SECTION = 'overload'
    RC_STOP_FROM_WEB = 8

    def __init__(self, core):
        super(Plugin, self).__init__(core)
        self.locks_list_dict = {}
        self.api_client = OverloadClient()
        self.jobno = None
        self.operator = ''
        self.retcode = -1
        self.copy_config = None
        self.jobno_file = None
        self.target = None
        self.lock_target_duration = None
        self.locks_list_cfg = None
        self.task = None
        self.job_name = None
        self.job_dsc = None
        self.notify_list = None
        self.version_tested = None
        self.regression_component = None
        self.is_regression = None
        self.ignore_target_lock = None
        self.port = None
        self.mon = None

    @staticmethod
    def get_key():
        return __file__

    def get_available_options(self):
        opts = [
            "api_address",
            "task",
            "job_name",
            "job_dsc",
            "notify",
            "ver",
        ]
        opts += [
            "component",
            "regress",
            "operator",
            "copy_config_to",
            "jobno_file",
        ]
        opts += ["token_file"]
        return opts

    @staticmethod
    def read_token(filename):
        if filename:
            logger.debug("Trying to read token from %s", filename)
            try:
                with open(filename, 'r') as handle:
                    data = handle.read().strip()
                    logger.info(
                        "Read authentication token from %s, "
                        "token length is %d bytes", filename, len(str(data)))
            except IOError:
                logger.error(
                    "Failed to read Overload API token from %s", filename)
                logger.info(
                    "Get your Overload API token from https://overload.yandex.net and provide it via 'overload.token_file' parameter"
                )
                raise RuntimeError("API token error")
            return data
        else:
            logger.error("Overload API token filename is not defined")
            logger.info(
                "Get your Overload API token from https://overload.yandex.net and provide it via 'overload.token_file' parameter"
            )
            raise RuntimeError("API token error")

    def configure(self):
        aggregator = self.core.job.aggregator_plugin
        aggregator.add_result_listener(self)

        self.api_client.set_api_address(self.get_option("api_address"))
        self.api_client.set_api_timeout(self.get_option("api_timeout", 30))
        self.api_client.set_api_token(
            self.read_token(self.get_option("token_file", "")))
        self.task = self.get_option("task", "DEFAULT")
        self.job_name = unicode(
            self.get_option("job_name", "none").decode("utf8"))
        if self.job_name == "ask" and sys.stdin.isatty():
            self.job_name = unicode(
                raw_input("Please, enter job_name: ").decode("utf8"))
        self.job_dsc = unicode(self.get_option("job_dsc", "").decode("utf8"))
        if self.job_dsc == "ask" and sys.stdin.isatty():
            self.job_dsc = unicode(
                raw_input("Please, enter job_dsc: ").decode("utf8"))
        self.notify_list = self.get_option("notify", "").split(" ")
        self.version_tested = unicode(self.get_option("ver", ""))
        self.regression_component = unicode(self.get_option("component", ""))
        self.is_regression = self.get_option("regress", "0")
        self.operator = self.get_option("operator", self.operator)
        if not self.operator:
            try:
                # Clouds and some virtual envs may fail this
                self.operator = pwd.getpwuid(os.geteuid())[0]
            except:
                logger.warning('failed to getpwuid.', exc_into=True)
                self.operator = 'unknown'
        self.copy_config = self.get_option("copy_config_to", "")
        self.jobno_file = self.get_option("jobno_file", "")

        if self.core.job.monitoring_plugin:
            self.mon = self.core.job.monitoring_plugin
            if self.mon.monitoring:
                self.mon.monitoring.add_listener(self)

        self.__save_conf()

    def prepare_test(self):
        try:
            console = self.core.get_plugin_of_type(ConsolePlugin)
        except KeyError:
            logger.debug("Console plugin not found", exc_info=True)
            console = None

        if console:
            console.add_info_widget(JobInfoWidget(self))
            console.remote_translator = self

        try:
            phantom = self.core.get_plugin_of_type(PhantomPlugin)
            info = phantom.get_info()
            self.target = info.address
        except (KeyError, AttributeError) as ex:
            logger.debug("No phantom plugin to get target info: %s", ex)
            self.target = socket.getfqdn()

        self.__save_conf()

    def start_test(self):
        try:
            phantom = self.core.get_plugin_of_type(PhantomPlugin)
            info = phantom.get_info()
            self.target = info.address
            port = info.port
            instances = info.instances
            tank_type = 1 if info.tank_type == "http" else 2
            # FIXME why don't we use resource_opener here?
            if info.ammo_file.startswith(
                    "http://") or info.ammo_file.startswith("https://"):
                ammo_path = info.ammo_file
            else:
                ammo_path = os.path.realpath(info.ammo_file)
            loadscheme = [] if isinstance(
                info.rps_schedule, unicode) else info.rps_schedule
            loop_count = info.loop_count
        except (KeyError, AttributeError) as ex:
            logger.debug("No phantom plugin to get target info: %s", ex)
            self.target = socket.getfqdn()
            port = 80
            instances = 1
            tank_type = 1
            ammo_path = ''
            loadscheme = []
            loop_count = 0

        try:
            jmeter = self.core.get_plugin_of_type(JMeterPlugin)
            ammo_path = jmeter.original_jmx
        except KeyError as ex:
            logger.debug("No jmeter plugin to get info: %s", ex)

        # try:
        #     pandora = self.core.get_plugin_of_type(PandoraPlugin)
        #     # TODO: get info from Pandora here
        # except KeyError as ex:
        #     logger.debug("No pandora plugin to get info: %s", ex)

        detailed_field = "interval_real"

        logger.info("Detected target: %s", self.target)

        self.jobno = self.api_client.new_job(
            self.task, self.operator,
            socket.getfqdn(), self.target, port, loadscheme, detailed_field,
            self.notify_list)
        web_link = "%s%s" % (self.api_client.address, self.jobno)
        logger.info("Web link: %s", web_link)
        self.publish("jobno", self.jobno)
        self.publish("web_link", web_link)
        self.make_symlink(self.jobno)
        self.set_option("jobno", str(self.jobno))
        if self.jobno_file:
            logger.debug("Saving jobno to: %s", self.jobno_file)
            fdes = open(self.jobno_file, 'w')
            fdes.write(str(self.jobno))
            fdes.close()

        self.api_client.edit_job_metainfo(
            self.jobno, self.job_name, self.job_dsc, instances, ammo_path,
            loop_count, self.version_tested, self.is_regression,
            self.regression_component, tank_type, " ".join(sys.argv), 0)

        self.__save_conf()

    def is_test_finished(self):
        return self.retcode

    def end_test(self, retcode):
        self.__save_conf()
        return retcode

    def post_process(self, rc):
        if self.jobno:
            try:
                self.api_client.close_job(self.jobno, rc)
            except Exception:  # pylint: disable=W0703
                logger.warning("Failed to close job", exc_info=True)

            logger.info("Web link: %s%s", self.api_client.address, self.jobno)

            autostop = None
            try:
                autostop = self.core.get_plugin_of_type(AutostopPlugin)
            except KeyError:
                logger.debug("No autostop plugin loaded", exc_info=True)

            if autostop and autostop.cause_criterion:
                rps = 0
                if autostop.cause_criterion.cause_second:
                    rps = autostop.cause_criterion.cause_second[1]["metrics"][
                        "reqps"]
                    if not rps:
                        rps = autostop.cause_criterion.cause_second[0][
                            "overall"]["interval_real"]["len"]
                self.api_client.set_imbalance_and_dsc(
                    self.jobno, rps, autostop.cause_criterion.explain())

            else:
                logger.debug("No autostop cause detected")
        self.__save_conf()
        return rc

    def __send_data(self, data_item, stat_item):
        if self.retcode < 0 and not self.api_client.push_test_data(
                self.jobno, data_item, stat_item):
            logger.warn("The test was stopped from Web interface")
            self.retcode = self.RC_STOP_FROM_WEB

    def on_aggregated_data(self, data, stats):
        """
        @data: aggregated data
        @stats: stats about gun
        """
        if not self.jobno:
            logger.warning("No jobNo gained yet")
            return
        self.__send_data(data, stats)

    def monitoring_data(self, data_list):
        if not self.jobno:
            logger.debug("No jobNo gained yet")
            return

        if self.retcode < 0:
            if "Telegraf" in self.core.job.monitoring_plugin.__module__:
                self.api_client.push_monitoring_data(
                    self.jobno, json.dumps(data_list))
            elif "Monitoring" in self.core.job.monitoring_plugin.__module__:
                [
                    self.api_client.push_monitoring_data(self.jobno, data)
                    for data in data_list if data
                ]
        else:
            logger.warn("The test was stopped from Web interface")

    def __save_conf(self):
        if self.copy_config:
            self.core.config.flush(self.copy_config)

        config = copy.copy(self.core.config.config)

        try:
            mon = self.core.get_plugin_of_type(MonitoringPlugin)
            config_filename = mon.config
            if config_filename and config_filename not in ['none', 'auto']:
                with open(config_filename) as config_file:
                    config.set(
                        MonitoringPlugin.SECTION, "config_contents",
                        config_file.read())
        except Exception:  # pylint: disable=W0703
            logger.debug("Can't get monitoring config", exc_info=True)

        output = StringIO.StringIO()
        config.write(output)
        if self.jobno:
            try:
                self.api_client.send_config_snapshot(
                    self.jobno, output.getvalue())
            except Exception:  # pylint: disable=W0703
                logger.debug("Can't send config snapshot: %s", exc_info=True)

    def send_console(self, text):
        try:
            self.api_client.send_console(self.jobno, text)
        except Exception:  # pylint: disable=W0703
            logger.debug("Can't send console snapshot: %s", exc_info=True)

    def make_symlink(self, name):
        PLUGIN_DIR = os.path.join(self.core.artifacts_base_dir, self.SECTION)
        if not os.path.exists(PLUGIN_DIR):
            os.makedirs(PLUGIN_DIR)
        os.symlink(self.core.artifacts_dir, os.path.join(PLUGIN_DIR, str(name)))

    def _core_with_tank_api(self):
        """
        Return True if we are running under Tank API
        """
        api_found = False
        try:
            import yandex_tank_api.worker  # pylint: disable=F0401
        except ImportError:
            logger.debug("Attempt to import yandex_tank_api.worker failed")
        else:
            api_found = isinstance(self.core, yandex_tank_api.worker.TankCore)
        logger.debug(
            "We are%s running under API server", ""
            if api_found else " likely not")
        return api_found


class JobInfoWidget(AbstractInfoWidget):
    def __init__(self, sender):
        AbstractInfoWidget.__init__(self)
        self.owner = sender

    def get_index(self):
        return 1

    def render(self, screen):
        template = "Author: " + screen.markup.RED + "%s" + \
            screen.markup.RESET + \
            "%s\n   Job: %s %s\n  Web: %s%s"
        data = (
            self.owner.operator[:1], self.owner.operator[1:], self.owner.jobno,
            self.owner.job_name, self.owner.api_client.address,
            self.owner.jobno)

        return template % data



''' Module that collects android device battery usage '''

import logging
import subprocess

from ...common.interfaces import AbstractPlugin

logger = logging.getLogger(__name__)


class Plugin(AbstractPlugin):
    """ Plugin that collects android device battery usage """
    SECTION = "battery_historian"

    @staticmethod
    def get_key():
        return __file__

    def __init__(self, core):
        AbstractPlugin.__init__(self, core)
        self.logfile = None
        self.default_target = None
        self.device_id = None
        self.cmds = {
            "enable_full_log":
            "adb %s shell dumpsys batterystats --enable full-wake-history",
            "disable_full_log":
            "adb %s shell dumpsys batterystats --disable full-wake-history",
            "reset": "adb %s shell dumpsys batterystats --reset",
            "dump": "adb %s shell dumpsys batterystats"
        }

    def get_available_options(self):
        return ["device_id"]

    def configure(self):
        self.device_id = self.get_option("device_id", None).strip()
        if self.device_id:
            modify = '-s {device_id}'.format(device_id=self.device_id)
            for key, value in self.cmds.iteritems():
                self.cmds[key] = value % modify
            self.logfile = self.core.mkstemp(".log", "battery_historian_")
            self.core.add_artifact_file(self.logfile)

    def prepare_test(self):
        if self.device_id:
            try:
                out = subprocess.check_output(
                    self.cmds['enable_full_log'], shell=True)
                logger.debug('Enabling full-log: %s', out)
                out = subprocess.check_output(self.cmds['reset'], shell=True)
                logger.debug('Reseting battery stats: %s', out)
            except subprocess.CalledProcessError:
                logger.error(
                    'Error trying to prepare battery historian plugin',
                    exc_info=True)

    def end_test(self, retcode):
        if self.device_id:
            try:
                logger.debug('dumping battery stats')
                dump = subprocess.Popen(
                    self.cmds['dump'], stdout=subprocess.PIPE,
                    shell=True).communicate()[0]
                out = subprocess.check_output(
                    self.cmds['disable_full_log'], shell=True)
                logger.debug('Disabling fulllog: %s', out)
                out = subprocess.check_output(self.cmds['reset'], shell=True)
                logger.debug('Battery stats reset: %s', out)
            except subprocess.CalledProcessError:
                logger.error(
                    'Error trying to collect battery historian plugin data',
                    exc_info=True)
            with open(self.logfile, 'w') as f:
                f.write(dump)
        return retcode

    def is_test_finished(self):
        return -1



""" Resource Opener tool """
import logging
import os
import requests
import gzip
import hashlib
import traceback
from contextlib import closing

logger = logging.getLogger(__name__)


class FormatDetector(object):
    """ Format Detector
    """

    def __init__(self):
        self.formats = {'gzip': (0, b'\x1f\x8b'), 'tar': (257, b'ustar\x0000')}

    def detect_format(self, header):
        for fmt, signature in self.formats.iteritems():
            if signature[1] == header[signature[0]:len(signature[1])]:
                return fmt


class ResourceManager(object):
    """ Resource opener manager.
        Use resource_filename and resource_string methods.
    """

    def __init__(self):
        self.path = None
        self.openers = {
            'http': ('http://', HttpOpener),
            'https': ('https://', HttpOpener)
        }

    def resource_filename(self, path):
        """
        Args:
            path: str, resource file url or resource file absolute/relative path.

        Returns:
            string, resource absolute path (downloads the url to /tmp)
        """
        return self.get_opener(path).get_filename

    def resource_string(self, path):
        """
        Args:
            path: str, resource file url or resource file absolute/relative path.

        Returns:
            string, file content
        """
        opener = self.get_opener(path)
        filename = opener.get_filename
        try:
            size = os.path.getsize(filename)
            if size > 50 * 1024 * 1024:
                logger.warning(
                    'Reading large resource to memory: %s. Size: %s bytes',
                    filename, size)
        except Exception as exc:
            logger.debug('Unable to check resource size %s. %s', filename, exc)
        with opener(filename, 'r') as resource:
            content = resource.read()
        return content

    def get_opener(self, path):
        """
        Args:
            path: str, resource file url or resource file absolute/relative path.

        Returns:
            file object
        """
        self.path = path
        opener = None
        for opener_name, signature in self.openers.items():
            if self.path.startswith(signature[0]):
                opener = signature[1](self.path)
        if not opener:
            opener = FileOpener(self.path)
        return opener


class FileOpener(object):
    """ File opener.
    """

    def __init__(self, f_path):
        self.f_path = f_path
        self.fmt_detector = FormatDetector()

    def __call__(self):
        with open(self.f_path, 'rb') as resource:
            header = resource.read(300)
        fmt = self.fmt_detector.detect_format(header)
        logger.debug('Resource %s format detected: %s.', self.f_path, fmt)
        if fmt == 'gzip':
            return gzip.open(self.f_path, 'rb')
        else:
            return open(self.f_path, 'rb')

    @property
    def get_filename(self):
        return self.f_path

    @property
    def hash(self):
        hashed_str = os.path.realpath(self.f_path)
        stat = os.stat(self.f_path)
        cnt = 0
        for stat_option in stat:
            if cnt == 7:  # skip access time
                continue
            cnt += 1
            hashed_str += ";" + str(stat_option)
        hashed_str += ";" + str(os.path.getmtime(self.f_path))
        return hashed_str

    @property
    def data_length(self):
        return os.path.getsize(self.f_path)


class HttpOpener(object):
    """ Http url opener.
        Downloads small files.
        For large files returns wrapped http stream.
    """

    def __init__(self, url):
        self.url = url
        self.fmt_detector = FormatDetector()
        self.force_download = None
        self.data_info = None
        self.timeout = 10
        self.get_request_info()

    def __call__(self, *args, **kwargs):
        return self.open(*args, **kwargs)

    def open(self, *args, **kwargs):
        with closing(
                requests.get(
                    self.url, stream=True, verify=False,
                    timeout=self.timeout)) as stream:
            stream_iterator = stream.raw.stream(100, decode_content=True)
            header = stream_iterator.next()
            fmt = self.fmt_detector.detect_format(header)
            logger.debug('Resource %s format detected: %s.', self.url, fmt)
        if not self.force_download and fmt != 'gzip' and self.data_length > 10**8:
            logger.info(
                "Resource data is not gzipped and larger than 100MB. Reading from stream.."
            )
            return HttpStreamWrapper(self.url)
        else:
            downloaded_f_path = self.download_file()
            if fmt == 'gzip':
                return gzip.open(downloaded_f_path, mode='rb')
            else:
                return open(downloaded_f_path, 'rb')

    def download_file(self):
        hasher = hashlib.md5()
        hasher.update(self.hash)
        tmpfile_path = "/tmp/%s" % hasher.hexdigest()
        if os.path.exists(tmpfile_path):
            logger.info(
                "Resource %s has already been downloaded to %s . Using it..",
                self.url, tmpfile_path)
        else:
            logger.info("Downloading resource %s to %s", self.url, tmpfile_path)
            try:
                data = requests.get(self.url, verify=False, timeout=10)
            except requests.exceptions.Timeout as exc:
                raise RuntimeError(
                    'Connection timeout reached '
                    'trying to download resource: %s \n'
                    'via HttpOpener: %s' % (self.url, exc))
            f = open(tmpfile_path, "wb")
            f.write(data.content)
            f.close()
            logger.info(
                "Successfully downloaded resource %s to %s", self.url,
                tmpfile_path)
        return tmpfile_path

    def get_request_info(self):
        logger.info('Trying to get info about resource %s', self.url)
        req = requests.Request(
            'HEAD', self.url, headers={'Accept-Encoding': 'identity'})
        session = requests.Session()
        prepared = session.prepare_request(req)
        try:
            self.data_info = session.send(
                prepared,
                verify=False,
                allow_redirects=True,
                timeout=self.timeout)
        except (
                requests.exceptions.Timeout,
                requests.exceptions.ConnectionError) as exc:
            logger.warning(
                'Connection error trying to get info about resource %s \n'
                'Exception: %s \n'
                'Retrying...' % (self.url, exc))
            try:
                self.data_info = session.send(
                    prepared,
                    verify=False,
                    allow_redirects=True,
                    timeout=self.timeout)
            except Exception as exc:
                logger.debug(
                    'Connection error trying to get info about resource %s \n'
                    'Traceback:  %s' % (traceback.format_exc(exc), self.url))
                raise RuntimeError(
                    'Connection error trying to get info about resource %s.'
                    'Exception: %s' % (self.url, exc))
        finally:
            session.close()
        try:
            self.data_info.raise_for_status()
        except requests.exceptions.HTTPError as exc:
            if exc.response.status_code == 405:
                logger.info(
                    "Resource storage does not support HEAD method. Ignore proto error and force download file."
                )
                self.force_download = True
            else:
                raise RuntimeError(
                    'Invalid HTTP response '
                    'trying to get info about resource: %s \n'
                    'via HttpOpener: %s' % (self.url, exc))

    @property
    def get_filename(self):
        config_f_path = self.download_file()
        return config_f_path

    @property
    def hash(self):
        last_modified = self.data_info.headers.get("Last-Modified", '')
        return self.url + "|" + last_modified

    @property
    def data_length(self):
        data_length = int(self.data_info.headers.get("Content-Length", 0))
        return data_length


class HttpStreamWrapper:
    """
    makes http stream to look like file object
    """

    def __init__(self, url):
        self.url = url
        self.buffer = ''
        self.pointer = 0
        self.stream_iterator = None
        self._content_consumed = False
        self.chunk_size = 10**3
        try:
            self.stream = requests.get(
                self.url, stream=True, verify=False, timeout=10)
            self.stream_iterator = self.stream.iter_content(self.chunk_size)
        except (
                requests.exceptions.Timeout,
                requests.exceptions.ConnectionError) as exc:
            raise RuntimeError(
                'Connection errors or timeout reached '
                'trying to make stream while downloading resource: %s \n'
                'via HttpStreamWrapper: %s' % (self.url, exc))
        try:
            self.stream.raise_for_status()
        except requests.exceptions.HTTPError as exc:
            raise RuntimeError(
                'Invalid HTTP response'
                'trying to open stream for resource: %s\n'
                'via HttpStreamWrapper: %s' % (self.url, exc))

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.stream.connection.close()

    def __iter__(self):
        while True:
            yield self.next()

    def _reopen_stream(self):
        self.stream.connection.close()
        try:
            self.stream = requests.get(
                self.url, stream=True, verify=False, timeout=30)
            self.stream_iterator = self.stream.iter_content(self.chunk_size)
        except (
                requests.exceptions.Timeout,
                requests.exceptions.ConnectionError) as exc:
            raise RuntimeError(
                'Connection errors or timeout reached '
                'trying to reopen stream while downloading resource: %s \n'
                'via HttpStreamWrapper: %s' % (self.url, exc))
        try:
            self.stream.raise_for_status()
        except requests.exceptions.HTTPError as exc:
            raise RuntimeError(
                'Invalid HTTP response'
                'trying to reopen stream for resource: %s\n'
                'via HttpStreamWrapper: %s' % (self.url, exc))
        self._content_consumed = False

    def _enhance_buffer(self):
        self.buffer += self.stream_iterator.next()

    def tell(self):
        return self.pointer

    def seek(self, position):
        if self.pointer:
            self.buffer = ''
            self._reopen_stream()
            self._enhance_buffer()
            while len(self.buffer) < position:
                self._enhance_buffer()
            self.pointer = position
            self.buffer = self.buffer[position:]

    def next(self):
        while '\n' not in self.buffer:
            try:
                self._enhance_buffer()
            except (
                    StopIteration, TypeError,
                    requests.exceptions.StreamConsumedError):
                self._content_consumed = True
                break
        if not self._content_consumed or self.buffer:
            try:
                line = self.buffer[:self.buffer.index('\n') + 1]
            except ValueError:
                line = self.buffer
            self.pointer += len(line)
            self.buffer = self.buffer[len(line):]
            return line
        raise StopIteration

    def read(self, chunk_size):
        while len(self.buffer) < chunk_size:
            try:
                self._enhance_buffer()
            except (
                    StopIteration, TypeError,
                    requests.exceptions.StreamConsumedError):
                break
        if len(self.buffer) > chunk_size:
            chunk = self.buffer[:chunk_size]
        else:
            chunk = self.buffer
        self.pointer += len(chunk)
        self.buffer = self.buffer[len(chunk):]
        return chunk

    def readline(self):
        """
        requests iter_lines() uses splitlines() thus losing '\r\n'
        we need a different behavior for AmmoFileReader
        and we have to use our buffer because we have probably read
        a bunch into it already
        """
        try:
            return self.next()
        except StopIteration:
            return ''


manager = ResourceManager()



from setuptools import setup, find_packages

setup(
    name='yandextank',
    version='1.8.30',
    description='a performance measurement tool',
    longer_description='''
Yandex.Tank is a performance measurement and load testing automatization tool.
It uses other load generators such as JMeter, ab or phantom inside of it for
load generation and provides a common configuration system for them and
analytic tools for the results they produce.
''',
    maintainer='Alexey Lavrenuke (load testing)',
    maintainer_email='direvius@yandex-team.ru',
    url='http://yandex.github.io/yandex-tank/',
    namespace_packages=["yandextank", "yandextank.plugins"],
    packages=find_packages(exclude=["tests", "tmp", "docs", "data"]),
    install_requires=[
        'psutil>=1.2.1', 'requests>=2.5.1', 'paramiko>=1.16.0',
        'pandas>=0.18.0', 'numpy>=1.11.0', 'future>=0.16.0',
        'pip>=8.1.2',
        'matplotlib>=1.5.3', 'seaborn>=0.7.1',
    ],
    setup_requires=[
        'pytest-runner', 'flake8',
    ],
    tests_require=[
        'pytest',
    ],
    license='LGPLv2',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
        'Operating System :: POSIX',
        'Topic :: Software Development :: Quality Assurance',
        'Topic :: Software Development :: Testing',
        'Topic :: Software Development :: Testing :: Traffic Generation',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ],
    entry_points={
        'console_scripts': [
            'yandex-tank = yandextank.core.cli:main',
            'yandex-tank-check-ssh = yandextank.common.util:check_ssh_connection',
        ],
    },
    package_data={
        'yandextank.core': ['config/*'],
        'yandextank.api': ['config/*'],
        'yandextank.plugins.GraphiteUploader': ['config/*'],
        'yandextank.plugins.JMeter': ['config/*'],
        'yandextank.plugins.Monitoring': ['config/*'],
        'yandextank.plugins.Phantom': ['config/*'],
        'yandextank.plugins.TipsAndTricks': ['config/*'],
        'yandextank.plugins.Pandora': ['config/*'],
        'yandextank.plugins.Aggregator': ['config/*'],
        'yandextank.plugins.Telegraf': ['config/*'],
    },
    use_2to3=False, )



"""Target monitoring via SSH"""
import base64
import getpass
import logging
import os.path
import re
import tempfile
import time
from collections import defaultdict
from xml.etree import ElementTree as etree

from ...common.util import SecuredShell
from ...common.interfaces import MonitoringDataListener
import sys
if sys.version_info[0] < 3:
    import ConfigParser
else:
    import configparser as ConfigParser

logger = logging.getLogger(__name__)
logging.getLogger("paramiko.transport").setLevel(logging.WARNING)


def parse_xml(config):
    if os.path.exists(config):
        return etree.parse(config)
    else:
        return etree.fromstring(config)


class Config(object):
    """Config reader helper"""

    def __init__(self, config):
        self.tree = parse_xml(config)

    def loglevel(self):
        """Get log level from config file. Possible values: info, debug"""
        log_level = 'info'
        log_level_raw = self.tree.getroot().get('loglevel')
        if log_level_raw in ('info', 'debug'):
            log_level = log_level_raw
        return log_level


class AgentClient(object):
    """Agent client connection"""

    def __init__(self, adr, timeout):
        self.run = []
        self.host = adr['host']
        self.username = adr['username']
        self.python = adr['python']
        self.metric = adr['metric']
        self.port = adr['port']
        self.interval = adr['interval']
        self.custom = adr['custom']
        self.startups = adr['startups']
        self.shutdowns = adr['shutdowns']
        self.session = None
        self.buffer = ""
        self.ssh = SecuredShell(self.host, self.port, self.username, timeout)

        handle, cfg_path = tempfile.mkstemp('.cfg', 'agent_')
        os.close(handle)
        self.path = {
            # Destination path on remote host
            'AGENT_REMOTE_FOLDER': '/var/tmp/lunapark_monitoring',

            # Source path on tank
            'AGENT_LOCAL_FOLDER': os.path.dirname(__file__) + '/agent',
            'METRIC_LOCAL_FOLDER': os.path.dirname(__file__) + '/agent/metric',

            # Temp config path
            'TEMP_CONFIG': cfg_path
        }

    def start(self):
        """Start remote agent"""
        logger.debug('Start monitoring: %s', self.host)
        self.session = self.ssh.async_session(
            " ".join([
                "DEBUG=1", self.python, self.path['AGENT_REMOTE_FOLDER'] +
                '/agent.py', '-c', self.path['AGENT_REMOTE_FOLDER'] +
                '/agent.cfg', '-t', str(int(time.time()))
            ]))
        return self.session

    def read_maybe(self):
        chunk = self.session.read_maybe()
        if chunk:
            parts = chunk.rsplit('\n', 1)
            if len(parts) > 1:
                ready_chunk = self.buffer + parts[0] + '\n'
                self.buffer = parts[1]
                return ready_chunk
            else:
                self.buffer += parts[0]
                return None
        return None

    def create_agent_config(self, loglevel):
        """Creating config"""
        try:
            float(self.interval)
        except:
            raise ValueError(
                "Monitoring interval should be a number: '%s'" % self.interval)

        cfg = ConfigParser.ConfigParser()
        cfg.add_section('main')
        cfg.set('main', 'interval', self.interval)
        cfg.set('main', 'host', self.host)
        cfg.set('main', 'loglevel', loglevel)
        cfg.set('main', 'username', self.username)

        cfg.add_section('metric')
        cfg.set('metric', 'names', self.metric)

        cfg.add_section('custom')
        for method in self.custom:
            if self.custom[method]:
                cfg.set('custom', method, ','.join(self.custom[method]))

        cfg.add_section('startup')
        for idx, cmd in enumerate(self.startups):
            cfg.set('startup', "cmd%s" % idx, cmd)

        cfg.add_section('shutdown')
        for idx, cmd in enumerate(self.shutdowns):
            cfg.set('shutdown', "cmd%s" % idx, cmd)

        with open(self.path['TEMP_CONFIG'], 'w') as fds:
            cfg.write(fds)

        return self.path['TEMP_CONFIG']

    def install(self, loglevel):
        """Create folder and copy agent and metrics scripts to remote host"""
        logger.info(
            "Installing monitoring agent at %s@%s...", self.username, self.host)

        # create remote temp dir
        cmd = self.python + ' -c "import tempfile; print tempfile.mkdtemp();"'
        logger.info("Creating temp dir on %s", self.host)
        try:
            out, errors, err_code = self.ssh.execute(cmd)
        except:
            logger.error(
                "Failed to install monitoring agent to %s",
                self.host,
                exc_info=True)
            return None
        if errors:
            logging.error("[%s] error: '%s'", self.host, errors)
            return None

        if err_code:
            logging.error(
                "Failed to create remote dir via SSH"
                " at %s@%s, code %s: %s" %
                (self.username, self.host, err_code, out.strip()))
            return None

        remote_dir = out.strip()
        if remote_dir:
            self.path['AGENT_REMOTE_FOLDER'] = remote_dir
        logger.debug(
            "Remote dir at %s:%s", self.host, self.path['AGENT_REMOTE_FOLDER'])

        # Copy agent and config
        agent_config = self.create_agent_config(loglevel)
        try:
            self.ssh.send_file(
                self.path['AGENT_LOCAL_FOLDER'] + '/agent.py',
                self.path['AGENT_REMOTE_FOLDER'] + '/agent.py')
            self.ssh.send_file(
                agent_config, self.path['AGENT_REMOTE_FOLDER'] + '/agent.cfg')
        except:
            logger.error(
                "Failed to install agent on %s", self.host, exc_info=True)
            return None
        return agent_config

    def uninstall(self):
        """
        Remove agent's files from remote host
        """
        if self.session:
            self.session.send("stop\n")
            self.session.close()
        fhandle, log_filename = tempfile.mkstemp(
            '.log', "agent_" + self.host + "_")
        os.close(fhandle)
        try:
            self.ssh.get_file(
                self.path['AGENT_REMOTE_FOLDER'] + "_agent.log", log_filename)
            self.ssh.rm_r(self.path['AGENT_REMOTE_FOLDER'])
        except:
            logger.error("Exception while uninstalling agent", exc_info=True)

        logger.info("Removing agent from: %s@%s...", self.username, self.host)
        return log_filename


class MonitoringCollector(object):
    """Aggregate data from several collectors"""

    def __init__(self):
        self.config = None
        self.default_target = None
        self.agents = []
        self.agent_sessions = []
        self.filter_conf = {}
        self.listeners = []
        self.first_data_received = False
        self.send_data = []
        self.artifact_files = []
        self.inputs, self.outputs, self.excepts = [], [], []
        self.filter_mask = defaultdict(str)
        self.ssh_timeout = 5
        self.load_start_time = None

    def add_listener(self, obj):
        self.listeners.append(obj)

    def prepare(self):
        """Prepare for monitoring - install agents etc"""

        # Parse config
        agent_config = []
        if self.config:
            [agent_config, self.filter_conf] = self.getconfig(
                self.config, self.default_target)
        loglevel = Config(self.config).loglevel()

        logger.debug("filter_conf: %s", self.filter_conf)

        # Filtering
        for host in self.filter_conf:
            self.filter_mask[host] = []
        logger.debug("Filter mask: %s", self.filter_mask)

        # Creating agent for hosts
        logger.debug('Creating agents')
        for adr in agent_config:
            logger.debug('Creating agent: %s', adr)
            agent = AgentClient(adr, timeout=self.ssh_timeout)
            logger.debug('Install monitoring agent. Host: %s', agent.host)
            agent_config = agent.install(loglevel)
            if agent_config:
                self.agents.append(agent)
                self.artifact_files.append(agent_config)

    def start(self):
        """Start N parallel agents"""
        [agent.start() for agent in self.agents]

    def poll(self):
        """Poll agents for data"""
        for agent in self.agents:
            block = agent.read_maybe()
            if not block:
                continue
            lines = block.split("\n")

            for data in lines:
                logger.debug("Got data from agent: %s", data.strip())
                self.send_data.append(
                    self.filter_unused_data(
                        self.filter_conf, self.filter_mask, data))
                logger.debug("Data after filtering: %s", self.send_data)

        if not self.first_data_received and self.send_data:
            self.first_data_received = True
            logger.info("Monitoring received first data")
        else:
            self.send_collected_data()

        return len(self.outputs)

    def stop(self):
        """Shutdown agents"""
        logger.debug("Uninstalling monitoring agents")
        for agent in self.agents:
            self.artifact_files.append(agent.uninstall())

    def send_collected_data(self):
        """sends pending data set to listeners"""
        [
            listener.monitoring_data(self.send_data)
            for listener in self.listeners
        ]
        self.send_data = []

    def get_host_config(self, host, target_hint):

        default = {
            'System': 'csw,int',
            'CPU': 'user,system,iowait',
            'Memory': 'free,cached,used',
            'Disk': 'read,write',
            'Net': 'recv,send,rx,tx',
        }

        default_metric = ['CPU', 'Memory', 'Disk', 'Net']

        names = defaultdict()

        hostname = host.get('address').lower()
        if hostname == '[target]':
            if not target_hint:
                raise ValueError(
                    "Can't use [target] keyword with "
                    "no target parameter specified")
            logger.debug("Using target hint: %s", target_hint)
            hostname = target_hint.lower()
        stats = []
        startups = []
        shutdowns = []
        custom = {
            'tail': [],
            'call': [],
        }
        metrics_count = 0
        for metric in host:
            # known metrics
            if metric.tag in default.keys():
                metrics_count += 1
                metr_val = default[metric.tag].split(',')
                if metric.get('measure'):
                    metr_val = metric.get('measure').split(',')
                for elm in metr_val:
                    if not elm:
                        continue
                    stat = "%s_%s" % (metric.tag, elm)
                    stats.append(stat)
                    agent_name = self.get_agent_name(metric.tag, elm)
                    if agent_name:
                        names[agent_name] = 1
            # custom metric ('call' and 'tail' methods)
            elif (str(metric.tag)).lower() == 'custom':
                metrics_count += 1
                isdiff = metric.get('diff')
                if not isdiff:
                    isdiff = 0
                stat = "%s:%s:%s" % (
                    base64.b64encode(metric.get('label')),
                    base64.b64encode(metric.text), isdiff)
                stats.append('Custom:' + stat)
                custom[metric.get('measure', 'call')].append(stat)
            elif (str(metric.tag)).lower() == 'startup':
                startups.append(metric.text)
            elif (str(metric.tag)).lower() == 'shutdown':
                shutdowns.append(metric.text)

        logger.debug("Metrics count: %s", metrics_count)
        logger.debug("Host len: %s", len(host))
        logger.debug("keys: %s", host.attrib.keys())
        logger.debug("values: %s", host.attrib.values())

        # use default metrics for host
        if metrics_count == 0:
            for metric in default_metric:
                metr_val = default[metric].split(',')
                for elm in metr_val:
                    stat = "%s_%s" % (metric, elm)
                    stats.append(stat)
                    agent_name = self.get_agent_name(metric, elm)
                    if agent_name:
                        names[agent_name] = 1

        metric = ','.join(names.keys())
        if not metric and not custom:
            metric = "cpu-stat"

        return {
            'metric': metric,
            'interval': host.get('interval', 1),
            'priority': host.get('priority', 0),
            'port': int(host.get('port', 22)),
            'python': host.get('python', '/usr/bin/env python2'),
            'username': host.get('username', getpass.getuser()),
            'custom': custom,
            'host': hostname,
            'startups': startups,
            'shutdowns': shutdowns,

            # XXX: should be separate?
            'stats': {
                hostname: stats
            },
        }

    def getconfig(self, filename, target_hint):
        """Prepare config data"""

        try:
            tree = parse_xml(filename)
        except IOError as exc:
            logger.error("Error loading config: %s", exc)
            raise RuntimeError("Can't read monitoring config %s" % filename)

        hosts = tree.findall('Host')
        config = []

        filter_obj = defaultdict(str)
        for host in hosts:
            host_config = self.get_host_config(host, target_hint)
            # XXX: why stats should be separated?
            filter_obj.update(host_config.pop('stats'))
            config.append(host_config)

        return [config, filter_obj]

    def filtering(self, mask, filter_list):
        """Filtering helper"""
        host = filter_list[0]
        initial = [0, 1]
        res = []
        if mask[host]:
            keys = initial + mask[host]
            for key in keys:
                try:
                    res.append(filter_list[key])
                except IndexError:
                    logger.warn(
                        "Problems filtering data: %s with %s", mask,
                        len(filter_list))
                    return None
        return ';'.join(res)

    def filter_unused_data(self, filter_conf, filter_mask, data):
        """Filter unselected metrics from data"""
        logger.debug("Filtering data: %s", data)
        out = ''
        # Filtering data
        keys = data.rstrip().split(';')
        if re.match('^start;', data):  # make filter_conf mask
            host = keys[1]
            for i in range(3, len(keys)):
                if keys[i] in filter_conf[host]:
                    filter_mask[host].append(i - 1)
            logger.debug("Filter mask: %s", filter_mask)
            out = 'start;'
            out += self.filtering(filter_mask, keys[1:]).rstrip(';') + '\n'
        elif re.match('^\[debug\]', data):  # log debug output
            logger.debug('agent debug: %s', data.rstrip())
        else:
            # if we are in start_test() phase, check data's timestamp with load_start_time
            # and skip data collected before load_start_time
            if self.load_start_time is not None:
                try:
                    if int(keys[1]) >= self.load_start_time:
                        filtered = self.filtering(filter_mask, keys)
                        if filtered:
                            out = filtered + '\n'  # filtering values
                except IndexError:
                    pass
        return out

    def get_agent_name(self, metric, param):
        """Resolve metric name"""
        depend = {
            'CPU': {
                'idle': 'cpu-stat',
                'user': 'cpu-stat',
                'system': 'cpu-stat',
                'iowait': 'cpu-stat',
                'nice': 'cpu-stat'
            },
            'System': {
                'la1': 'cpu-la',
                'la5': 'cpu-la',
                'la15': 'cpu-la',
                'csw': 'cpu-stat',
                'int': 'cpu-stat',
                'numproc': 'cpu-stat',
                'numthreads': 'cpu-stat',
            },
            'Memory': {
                'free': 'mem',
                'used': 'mem',
                'cached': 'mem',
                'buff': 'mem',
            },
            'Disk': {
                'read': 'disk',
                'write': 'disk',
            },
            'Net': {
                'recv': 'net',
                'send': 'net',
                'tx': 'net-tx-rx',
                'rx': 'net-tx-rx',
                'retransmit': 'net-retrans',
                'estab': 'net-tcp',
                'closewait': 'net-tcp',
                'timewait': 'net-tcp',
            }
        }
        if depend[metric][param]:
            return depend[metric][param]
        else:
            return ''


class StdOutPrintMon(MonitoringDataListener):
    """Simple listener, writing data to stdout"""

    def __init__(self):
        MonitoringDataListener.__init__(self)

    def monitoring_data(self, data_list):
        [sys.stdout.write(data) for data in data_list]


class MonitoringDataDecoder(object):
    """The class that serves converting monitoring data lines to dict"""
    NA = 'n/a'

    def __init__(self):
        self.metrics = {}

    def decode_line(self, line):
        """convert mon line to dict"""
        is_initial = False
        data_dict = {}
        data = line.strip().split(';')
        timestamp = -1
        if data[0] == 'start':
            data.pop(0)  # remove 'start'
            host = data.pop(0)
            if not data:
                logger.warn("Wrong mon data line: %s", line)
            else:
                timestamp = data.pop(0)
                self.metrics[host] = []
                for metric in data:
                    if metric.startswith("Custom:"):
                        metric = base64.standard_b64decode(metric.split(':')[1])
                    self.metrics[host].append(metric)
                    data_dict[metric] = self.NA
                    is_initial = True
        else:
            host = data.pop(0)
            timestamp = data.pop(0)

            if host not in self.metrics.keys():
                raise ValueError(
                    "Host %s not in started metrics: %s" % (host, self.metrics))

            if len(self.metrics[host]) != len(data):
                raise ValueError(
                    "Metrics len and data len differs: %s vs %s" %
                    (len(self.metrics[host]), len(data)))

            for metric in self.metrics[host]:
                data_dict[metric] = data.pop(0)

        logger.debug("Decoded data %s: %s", host, data_dict)
        return host, data_dict, is_initial, timestamp


# FIXME: 3 synchronize times between agent and collector better
